use rand::prelude::*;

use std::{
    cell::{Ref, RefCell},
    collections::HashMap,
    rc::{Rc, Weak},
};

use crate::prelude::*;

use fleck_core::crypto::{Keypair, PublicKey};

pub struct InnerRef<'l, T> {
    r: Ref<'l, Option<T>>,
}

impl<'l, T> InnerRef<'l, T> {
    fn build(from: Ref<'l, Option<T>>) -> Option<Self> {
        if from.is_none() {
            None
        } else {
            Some(Self { r: from })
        }
    }
}

impl<'l, T> AsRef<T> for InnerRef<'l, T> {
    fn as_ref(&self) -> &T {
        self.r.as_ref().unwrap()
    }
}

impl<'l, T> std::ops::Deref for InnerRef<'l, T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        self.r.as_ref().unwrap()
    }
}

#[derive(Debug, Default)]
pub struct Node {
    pubkey: PublicKey,
    peer: Option<fleck_core::peer::Peer>,
    keypair: RefCell<Option<Keypair>>,
}

impl Node {
    pub fn build_with_peer(pubkey: PublicKey, peer: fleck_core::peer::Peer) -> Self {
        Self {
            pubkey,
            peer: Some(peer),
            ..Default::default()
        }
    }

    pub fn build_ephemeral() -> Self {
        let mut csprng = StdRng::from_entropy();
        let kp = Keypair::generate(&mut csprng);

        Self {
            pubkey: kp.public,
            peer: None,
            keypair: RefCell::new(Some(kp)),
        }
    }

    pub fn build_with_keypair(kp: Keypair) -> Self {
        Self {
            pubkey: kp.public,
            peer: None,
            keypair: RefCell::new(Some(kp)),
        }
    }

    pub fn peer(&self) -> Option<fleck_core::peer::Peer> {
        self.peer.clone()
    }

    pub fn pubkey(&self) -> &PublicKey {
        &self.pubkey
    }
    pub(crate) fn keypair(&self) -> Option<InnerRef<'_, Keypair>> {
        InnerRef::build(self.keypair.borrow())
    }
}

pub struct RegTag {}
pub type NodeRegistrationChannel = (RegTag, Rc<Node>);

pub struct NodeService {
    api: Rc<crate::API>,
    all_nodes: RefCell<Vec<Rc<Node>>>,
    // node_by_pubkey: RefCell<HashMap<fleck_core::crypto::PublicKey, Rc<Node>>>,
    node_by_peer: RefCell<HashMap<fleck_core::Peer, Rc<Node>>>,
    self_node: RefCell<Weak<Node>>,
}

impl NodeService {
    pub fn self_node(&self) -> Rc<Node> {
        self.self_node
            .borrow()
            .upgrade()
            .expect("asking for self node before it's been initialized!")
    }

    pub fn build_ephemeral_self_node(&self) {
        assert_eq!(self.self_node.borrow().strong_count(), 0);

        let nself = Rc::new(Node::build_ephemeral());

        self.all_nodes.borrow_mut().push(nself.clone());
        *self.self_node.borrow_mut() = nself.as_weak();
    }

    pub fn node_by_pubkey(&self, pubkey: &PublicKey) -> Option<Rc<Node>> {
        self.all_nodes
            .borrow()
            .iter()
            .find(|n| n.pubkey() == pubkey)
            .map(|r| r.to_owned())
    }

    // nodes we believe we can communicate with directly
    pub fn direct_neighbours(&self) -> Vec<Rc<Node>> {
        self.all_nodes
            .borrow()
            .iter()
            .filter(|n| n.peer.is_some())
            .map(|r| r.to_owned())
            .collect()
    }

    pub fn inform_of(&self, pubkey: PublicKey, peer: fleck_core::Peer) {
        // do we know of a node with these properties already?
        let pref = &peer;
        log::trace!(
            "checking to see if we already know of peer: ({:?}, {:?})",
            pubkey,
            peer
        );
        for node in self.all_nodes.borrow().iter() {
            log::trace!("\toption: {:?}, {:?}", node.pubkey, node.peer);
        }
        if self
            .all_nodes
            .borrow()
            .iter()
            .filter(|n| n.peer.as_ref() == Some(pref) && n.pubkey == pubkey)
            .count()
            != 0
        {
            return;
        }

        // nope, it's a new node
        log::trace!("Informed of new peer!");
        let node = Rc::new(Node::build_with_peer(pubkey, peer.clone()));

        self.all_nodes.borrow_mut().push(node.clone());
        self.node_by_peer.borrow_mut().insert(peer, node.clone());
        self.api.channel::<NodeRegistrationChannel>().queue(node);
    }
}

impl Service for NodeService {
    fn new(api: Rc<crate::API>) -> Self {
        Self {
            api,
            all_nodes: Default::default(),
            // node_by_pubkey: Default::default(),
            node_by_peer: Default::default(),
            self_node: Default::default(),
        }
    }

    fn setup(self: &Rc<Self>) {
        self.api.create_channel::<NodeRegistrationChannel>();

        self.api
            .channel::<fleck_core::ReceivePacketChannel>()
            .sub_ref(fleck_core::ReceiveOrder::Annotate, self, Self::annotate);
    }
}

impl NodeService {
    fn annotate(&self, msg: &mut fleck_core::Message) {
        // annotate node based on peer
        if let Some(peer) = &msg.peer {
            let map = self.node_by_peer.borrow();
            match map.get(peer) {
                Some(node) => {
                    msg.node = Some(node.clone());
                }
                None => {
                    // the interesting case, it's a new node!
                    // TODO: send them a request for their public key
                    log::warn!("incoming message from unknown source; peer is {:?}", &peer);
                }
            }
        }
    }
}
